#!/usr/bin/perl -w
#$Id: compute_robust_context.pl,v 1.11 2010/07/07 12:55:14 rnl14480 Exp $

use strict;
use lib qw(../lib);
use BAOBAB::Measure;
use BAOBAB::MeasureWrapper;
use BAOBAB::Evaluation;
use BAOBAB::EvaluationWrapper;
use BAOBAB::Precursor;
use BAOBAB::PrecursorWrapper;
use BAOBAB::Wrapper;
use BAOBAB::HMM;
use BAOBAB::Fasta;
use DBI;

use Data::Dumper qw(&Dumper);

print STDERR "Connecting to the database...";
my $dbh = DBI->connect(
	$BAOBAB::Wrapper::dsn, $BAOBAB::Wrapper::user,
	$BAOBAB::Wrapper::password, { RaiseError => 1, AutoCommit => 0 }
) || die "Could not connect to the database : $DBI::errstr\n";
$dbh->{mysql_auto_reconnect} = 1;

die "Usage: $0 data-dir [start [limit [extra]]]\n" unless scalar(@ARGV) >= 1;

my $data = shift;

my $start = shift;
my $limit = shift;
my $extra = shift;
$extra = undef unless $extra;

my $mw = BAOBAB::MeasureWrapper->new( dbh => $dbh );
my $m = $mw->read( name => 'robust_context' );

my %params = ();
if ( defined $start && defined $limit ) {
	%params = ( start => $start, limit => $limit );
}
$params{extra} = $extra if defined $extra;

my $pw = BAOBAB::PrecursorWrapper->new( dbh => $dbh );
my $pi = $pw->iterator(%params);

my %fasta = ();

while ( my $pid = $pi->() ) {
	my $p    = $pw->read( id => $pid, no_cache => 1 );
	my $org  = $p->chromosome->organism->shortname;
	my $file = $data . "$org.nfa";

	my $fobj;
	if ( exists $fasta{$org} ) {
		$fobj = $fasta{$org};
	}
	else {

		$fobj = BAOBAB::Fasta->new(
			filename       => $file,
			header_handler => get_handler($org),
			save_memory    => 1
		);
		$fasta{$org} = $fobj;
	}

	my ( $up, $down ) = ( 0, 0 );
	my ( $cstart, $cstop ) = ( $p->start, $p->stop );
	my $csize =
	  $p->stop - $p->start + 1;    # the context is as big as the sequence

	if ( $cstart - $csize >= 0 ) {
		$cstart -= $csize;
		$up = $csize;
	}
	else {
		$up     = $cstart - 0;
		$cstart = 0;
	}

	if ( $cstop + $csize <= $p->chromosome->size - 1 ) {
		$cstop += $csize;
		$down = $csize;
	}
	else {
		$down  = $p->chromosome->size - 1 - $cstop;
		$cstop = $p->chromosome->size - 1;
	}

	my $cseq = $fobj->get_region(
		chr    => $p->chromosome->name,
		offset => $cstart,
		len    => $cstop - $cstart + 1,
		strand => $p->strand
	);

	my $uhmm = BAOBAB::HMM->new();
	my $dhmm = BAOBAB::HMM->new();

	$uhmm->train( seqs => [ substr( $cseq, 0, $up ) ] );
	$dhmm->train( seqs => [ substr( $cseq, $up + $p->length, $down ) ] );

	my @dist = ();
	for ( my $i = 0 ; $i < 100 ; $i++ ) {

		my $ctxt =
		    $uhmm->generate( len => $up, rna => 1 )
		  . $p->sequence
		  . $dhmm->generate( len => $down, rna => 1 );

		my ( $str, undef ) = RNA::fold($ctxt);
		my $excise = substr $str, $up, $p->length;

		$excise = _repair_structure($excise);
		my $pres = _preserved_structure( $p->structure, $excise );
		push @dist, $pres;

	}

	@dist = sort { $a <=> $b } @dist;

	my $midpoint        = scalar(@dist) / 2;
	my $median_distance = 0;
	if ( $median_distance && scalar(@dist) % 2 ) {
		$median_distance = $dist[ int($midpoint) ];
	}
	else {
		$median_distance = ( $dist[$midpoint] + $dist[ $midpoint - 1 ] ) / 2;
	}

	my $value = $median_distance;

	my $e = BAOBAB::Evaluation->new(
		precursor => $p,
		measure   => $m,
		value     => $value,
		no_cache  => 1
	);

	my $ew = BAOBAB::EvaluationWrapper->new( dbh => $dbh, object => $e );
	$ew->write;

}

$dbh->disconnect;

sub get_handler {

	my $o    = shift;
	my %data = (
		ada => sub { $_ = shift; s/\s//g; return $_; },
		aga                    => sub { ( split /:/, shift )[1] },
		planted_metazoa        => sub { return 'U' },
		random_ada             => sub { return 'U' },
		random_aga             => sub { return 'U' },
		random_dmel            => sub { return 'U' },
		random_planted_metazoa => sub { return 'U' },
		dmel                   => sub {
			$_ = shift;
			if (/^arm_(\S+)/) {
				return $1;
			}
			elsif (/^Arm(\S+)/) {
				return $1;
			}
			elsif (/^(\S+)/) {
				return $1;
			}
			else {
				return '??';
			}
		}
	);

	return $data{$o};
}

sub _repair_structure {
	my $str  = shift;
	my @left = ();
	my @str  = split //, $str;

	my $i     = 0;
	my @stack = ();
	foreach my $p (@str) {
		if ( $p eq '(' ) {
			push @stack, $i;
		}
		elsif ( $p eq ')' ) {
			if (@stack) {
				pop(@stack);
			}
			else {
				$str[$i] = '.';
			}
		}
		$i++;
	}
	foreach my $r (@stack) {
		$str[$r] = '.';
	}

	return join( "", @str );
}

sub _preserved_structure {
	my $str1 = shift;
	my $str2 = shift;

	return 0 unless length($str1) > 0 and length($str2) > 0;
	return undef unless length($str1) == length($str2);
	my $len = length($str1);

	my $clen = 0;
	my $mlen = 0;
	foreach my $i ( 0 .. $len - 1 ) {
		if ( substr( $str1, $i, 1 ) eq substr( $str2, $i, 1 ) ) {
			$clen++;
		}
		else {
			$mlen = $clen if $clen > $mlen;
			$clen = 0;
		}
	}
	$mlen = $clen if $clen > $mlen;

	return $mlen / $len;
}
